const xmldoc = require('xmldoc')
const template = require('art-template')
require('./helper')(template)
const fs = require('fs');
const cache = require('memory-cache');
const ajax = require('mjax');
const mlog = require('mqlog');
const marr = require('marr');
const natis = require('./natis');
const mvalidate = require('mvalidate');
const Tool = require('../tool')

require('mstr');
const {toCamelCase} = require("echarts/lib/util/format");

let uuidLen = 8;
const pool = {};


function underlined(name){
    return name.replace(/\B([A-Z])/g, '_$1').toLowerCase()
}


function camel(str){
    return (str || '').replace(/_(.)/g, function (match, group1) {
        return group1.toUpperCase();
    })
}


function camelizeList(data){
    let all = []
    for(let i in data){
        let item = {}
        for(let key in data[i]) {
            let newKey = camel(key)
            item[newKey] = data[i][key]
        }
        all.push(item)
    }
    return all;
}

//给不同的connection设置不同的参数
!function(){
    let util = {
        config : function(config){
            let connName = config.jdbc.connName;
            pool[connName] = {};
            pool[connName]['jdbc']  = config.jdbc || {};
            pool[connName]['mapper'] = config.mapper;
            pool[connName]['validate'] = config.validate || {};
            uuidLen = config.jdbc.uuid_len || 8;
        },
        errs : {
            uniqueError : '唯一性错误',
            checkError : '删除时检查错误',
            insertError : '插入失败',
            updateError : '更新失败',
            deleteError : '删除失败'
        },
        getExecsql : function(dbModel,model,method,param,resolve,reject,res){
            //为每次查询生成一个id,便于排查问题
            dbModel._id = Date.now()
            const jdbc = pool[dbModel.connName]['jdbc'];
            const validate = pool[dbModel.connName]['validate'];
            const mapperDir = pool[dbModel.connName]['mapper'] ;
            const key = model + "." + method;
            //先读缓存
            let cacheSql = cache.get(key);
            if(jdbc.debug){
                mlog.info(dbModel._id, "读取模板:", key)
            }
            let presql = cacheSql;
            if(!cacheSql || jdbc.debug){
                let filename = underlined(model)
                let path = mapperDir + '/'+ filename+'.xml';
                if(!fs.existsSync(path)){
                    path = mapperDir + '/' + model + '.xml'
                    if(!fs.existsSync(path)){
                        mlog.error(path + "文件不存在");
                        return util.reject('系统错误',reject,res);
                    }
                }
                console.log("bbbb", path)
                const presqlfile = fs.readFileSync(path,{encoding:"utf-8"});
                const document = new xmldoc.XmlDocument(presqlfile);
                const methodEle = document.childWithAttribute("id",method)

                if(!methodEle || !methodEle.val){
                    mlog.error(model+"的"+method+"方法未定义");
                    return util.reject('系统错误',reject,res);
                }
                presql = document.childWithAttribute("id",method).val;
                //写入缓存
                cache.put(key,presql,1000 * 60 * 5);
            }
            if(jdbc.debug){
                mlog.info(dbModel._id, "模板SQL:", presql);
            }
            if(jdbc.debug){
                mlog.info(dbModel._id, "传入参数:", param);
            }
            let sql = '';
            try{
                 sql = template.compile(presql,{compress:true,escape:false})(param);
            }
            catch(err){
                mlog.error(dbModel._id, "SQL模板编译出错");
                return util.reject("系统错误",reject,res);
            }

            sql = sql.trim();
            //处理若干异常SQL
            //where and
            sql = sql.replace(/where\s+and/g, 'where ');
            sql = sql.replace(/where\s+order by/g, 'order by');
            //where )
            sql = sql.replace(/where\s+\)/g, ')')
            sql = sql.replace(/,\s+where/g, ' where')
            sql = sql.replace(/\s+where\s+limit\s+/g, ' limit ')
            //where结尾
            let mql = sql.toLowerCase()
            if(mql.lastIndexOf("where") == mql.length - 5){
                console.log("空where")
                sql = sql.slice(0, -5);
                if(mql.includes("update ") && mql.includes(" set ")){
                    return util.reject("参数缺失",reject,res);
                }
            }
            if(jdbc.debug){
                mlog.info(dbModel._id, "执行SQL:", sql);
            }
            return sql;
        },
        reject : function(e,reject,res){
            if(res){
                if(typeof res == "function"){
                    return res(e);
                }
                return ajax.failure(res,e);
            }
            return reject(e);
        },
        resolve : function(data,resolve,res){
            if(res){
                if(typeof res == "function"){
                    return res(null,data);
                }
                return ajax.success(res,data);
            }
            return resolve(data);
        },
        exec : function(dbModel,model,method,param,resolve,reject,res,ele,deal){

            /*前置操作
             1.如果是删除操作,则进行合法性判断
             2.如果是分页查询操作,则要把查询总量也要返回

             后置数据处理
             1.如果对于非列表查询,则将返回的数据进行处理
             2.如果一对多查询,那么在查询后要处理成一对多的形式

             */
            let beforeMethod = ele.attr.before;

            //判断必填字段
            let mustField = ele.attr.must || ''

            if(mustField){
                let fields = mustField.split(",").map(one => one.trim())
                for(let field of fields){
                    console.log('check field', field)
                    if(!param[field]) {
                        mlog.error(`/api/${model}/${method}出现参数缺失：${field}`);
                        return util.reject('参数缺失',reject,res);
                    }
                }

            }
            console.log(model, method, 'must', mustField)

            let args = arguments;
            if(beforeMethod){ //需要前置处理
                //先判断是否存在此方法
                let bean = (dbModel.models)[model];
                if(bean.hasOwnProperty(beforeMethod)){
                    bean[beforeMethod](param).then(function(){
                        //继续执行
                        util.justDeal.apply(util,args);
                    },function(err){
                        mlog.error("前置处理出现错误:");
                        mlog.error(err);
                        return util.reject(err,reject,res);
                    });
                }
                else{
                    mlog.error("前置方法未定义,忽略执行");
                    util.justDeal.apply(util,args);
                }
            }
            else{
                util.justDeal.apply(util,args);
            }
        },
        justDeal : function(dbModel,model,method,param,resolve,reject,res,ele,deal,times){
            times = times || 0;
            const jdbc = pool[dbModel.connName]['jdbc'];
            const validate = pool[dbModel.connName]['validate'];
            const mapperDir = pool[dbModel.connName]['mapper'];
            const sql = this.getExecsql(dbModel,model,method,param,resolve,reject,res);
            //取新连接
            const connName = dbModel.connName
            natis.getConnection(connName, function(db){
                db.querying = true
                db.query({sql : sql,timeout : 10000},function(err,data){
                    db.querying = false;
                    //查询失败
                    if(err){
                        //判断失败类型
                        if(err.message.startsWith("ER_PARSE_ERROR")){
                            mlog.error(dbModel._id,"SQL错误" + sql);
                            return util.reject("非法请求",reject,res);
                        }
                        else if(err.message.startsWith("ER_TRUNCATED_WRONG_VALUE_FOR_FIELD")){
                            mlog.error(dbModel._id,"SQL字段赋值错误" + err.message);
                            return util.reject("非法请求",reject,res);
                        }
                        times ++;
                        mlog.error(dbModel._id, err);
                        //重新获取连接，并重新操作
                        // mlog.info("准备重试...");
                        // // dbModel.connection.disabled = true;
                        // mlog.info(dbModel.connName);
                        //
                        // natis.getDb(dbModel.connName, function(db){
                        //     if(times == 3){
                        //         mlog.error(err);
                        //         if(jdbc.debug){
                        //             mlog.error("数据操作失败");
                        //             mlog.error(err);
                        //         }
                        //         return util.reject("服务器错误",reject,res);
                        //     }
                        //     util.justDeal.call(util, db,model,method,param,resolve,reject,res,ele,deal,times);
                        // });
                        return util.reject("非法请求", reject, res)
                    }

                    //调用回调方法处理MYSQL返回结果
                    try{
                        data = deal(data,method,ele);
                    }catch (e){
                        let msg = e.message;
                        switch (msg){
                            case util.errs.uniqueError :
                                if(validate[model] && validate[model].unique) {
                                    msg = validate[model].unique;
                                }
                                else if(ele.attr.msg){
                                    msg = ele.attr.msg;
                                }
                                else{
                                    msg = "存在相同记录";
                                }
                                break;
                            case util.errs.checkError:
                                if(validate[model] && validate[model].check) {
                                    msg = validate[model].check;
                                }
                                else if(ele.attr.msg){
                                    msg = ele.attr.msg;
                                }
                                else{
                                    msg = "该记录已被关联,无法删除";
                                }
                                break;
                            default :
                                break;
                        }
                        return util.reject(msg,reject,res);
                    }
                    if(jdbc.debugData){
                        mlog.info(dbModel._id,"处理结果:", data)
                    }
                    return util.resolve(data,resolve,res);
                });
            })


        },
        select : function(){
            util.exec.apply(util,util.getArgs(arguments,function(data,method,ele){
                if(data == null || data.length == 0){
                    //取单个值
                    if(ele.attr.single){
                        return null
                    }
                    return [];
                }
                //对数据进行驼峰化
                data = camelizeList(data)

                if(ele.attr.tree){
                    data = marr.generateTreeByPid(data,ele.attr.idKey,ele.attr.pidKey,ele.attr.pidValue);
                }
                else if(ele.attr.child){
                    data = marr.generateTreeByCatalog(data,ele.attr.child,ele.attr.key,ele.attr.childName);
                }

                if(ele.attr.one || method == 'get'){
                    data = data[0]
                }
                if(data.length == 1 && util.isCountSql(data[0])){
                    data = data[0]
                }
                //结果集只有一个元素
                if(!Array.isArray(data) && util.countEle(data) == 1 && (ele.attr.bool || ele.attr.one || util.isCountSql(data))){
                    data = util.getFirstEle(data);
                }

                //取单个值
                if(ele.attr.single){
                    data = data.flatMap(o => Object.values(o))
                }

                if(ele.attr.bool){
                    return data > 0;
                }

                if(method.startsWith("isUnique","unique")){
                    if(data > 0){
                        throw new Error(util.errs.uniqueError);
                    }
                }
                else if(method.startsWith("isCheck","check")){
                    if(data > 0){
                        throw new Error(util.errs.checkError);
                    }
                }
                return data;
            }));
        },
        insert : function(dbModel,model,method,param,resolve,reject,res,ele){

            // user : {  //key
            //     'name' : { // col
            //         minlen : {val:5,message:'用户名至少6位'},  //ruleItem = ruleValue
            //         require : {val:true,messagge:'用户名不能为空'} //ruleItem = ruleValue
            //     },
            // }
            //验证
            const jdbc = pool[dbModel.connName]['jdbc'];
            const validate = pool[dbModel.connName]['validate'];
            const mapperDir = pool[dbModel.connName]['mapper'] ;
            const validateModel = validate[model];

            if(validateModel){
                //验证
                if(jdbc.debug){
                    mvalidate.setDebug(true);
                }
                let validateOk = true;
                let msg = '';
                for(let col in validateModel){
                    let colVal = param[col]; //取值
                    let rules = validateModel[col]; //取规则
                    if(typeof rules == "string"){
                        continue;
                    }
                    let validateErrorItem = '';
                    for(let ruleItem in rules){
                        let ruleVal = rules[ruleItem]['val']; //取规则值
                        validateErrorItem = ruleItem;
                        if(ruleItem == "required" && ruleVal && (colVal == null || colVal == '')){
                            validateOk = false;
                            break;
                        }
                        if(ruleItem != "required" && !mvalidate.validate(ruleItem,colVal,ruleVal)){
                            validateOk = false;
                            break;
                        }
                    }
                    if(!validateOk){
                        if(rules[validateErrorItem]['message']){
                            msg = rules[validateErrorItem]['message'];
                            mlog.error("字段验证失败:"+ msg);
                        }
                        else{
                            mlog.error('缺少字段' + validateErrorItem + '验证失败提示信息配置');
                            msg = '字段信息校验失败';
                        }
                        return util.reject(msg,reject,res);
                    }
                }
            }

            if(!param.uuid){
                param.uuid = Tool.uuid()
            }
            util.exec.apply(util,util.getArgs(arguments,function(data,method,ele){
                let affected = data.affectedRows;
                if(affected>0){
                    return {
                        "uuid": param.uuid,
                        'id': data.insertId
                    }
                }
                else{
                    throw new Error(util.errs.insertError);
                }
            }));
        },
        update : function(dbModel,model,method,param,resolve,reject,res,ele){
            const jdbc = pool[dbModel.connName]['jdbc'];
            const validate = pool[dbModel.connName]['validate'];
            const mapperDir = pool[dbModel.connName]['mapper'] ;
            const validateModel = validate[model];

            if(validateModel){
                //验证
                if(jdbc.debug){
                    mvalidate.setDebug(true);
                }
                let validateOk = true;
                let msg = '';
                for(let col in validateModel){
                    let colVal = param[col]; //取值
                    let rules = validateModel[col]; //取规则
                    if(typeof rules == "string"){
                        continue;
                    }
                    let validateErrorItem = '';
                    for(let ruleItem in rules){
                        let ruleVal = rules[ruleItem]['val']; //取规则值
                        validateErrorItem = ruleItem;

                        if(ruleItem == "required" && ruleVal && (colVal == '')){
                            validateOk = false;
                            break;
                        }
                        if(ruleItem != "required" && !mvalidate.validate(ruleItem,colVal,ruleVal)){
                            validateOk = false;
                            break;
                        }
                    }
                    if(!validateOk){
                        if(rules[validateErrorItem]['message']){
                            msg = rules[validateErrorItem]['message'];
                            mlog.error("字段验证失败:"+ msg);
                        }
                        else{
                            mlog.error('缺少字段' + validateErrorItem + '验证失败提示信息配置');
                            msg = '字段信息校验失败';
                        }
                        return util.reject(msg,reject,res);
                    }
                }
            }


            util.exec.apply(util,util.getArgs(arguments,function(data,method,ele){
                let affected = data.affectedRows;
                if(affected>0){
                    return 1;
                }
                else{
                    return 0;
                }
            }));
        },
        delete : function(){
            util.exec.apply(util,util.getArgs(arguments,function(data,method,ele){
                let affected = data.affectedRows;
                if(affected>0){
                    return 1;
                }
                else{
                    return 0;
                }
            }));
        },
        countEle : function(o){
            if(typeof o == "object"){
                let sum = 0;
                for(let i in o){
                    sum ++;
                }
                return sum;
            }
            return -1;
        },
        getFirstEle : function(o){
            if(typeof o == "object"){
                for(let i in o){
                    return o[i];
                }
            }
            return -1;
        },
        isCountSql : function(o){
            for(let key in o){
                return key.toLowerCase().indexOf("count(") == 0;
            }
            return false;
        },
        getArgs : function(args,other1,other2){
            let arr = [];
            for(let i in args){
                arr[i] = args[i];
            }
            if(typeof other1 != "undefined"){
                arr.push(other1);
            }
            if(typeof other2 != "undefined"){
                arr.push(other2);
            }
            return arr;
        }
    }
    module.exports = util;

}();
